// SPDX-License-Identifier: GPL-2.0+
/* Copyright (c) 2006-2019 AVM GmbH <fritzbox_info@avm.de> */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <asm/fcntl.h>
#include <asm/ioctl.h>
#include <asm/errno.h>
#include <asm/uaccess.h>
#include <avm/watchdog/watchdog.h>
#include "avm_sammel.h"

/*-----------------------------------------------------------------------------------------------*\
\*-----------------------------------------------------------------------------------------------*/
/*--- #define AVM_WATCHDOG_DEBUG ---*/

#if defined(AVM_WATCHDOG_DEBUG)
#define DBG(...) printk(KERN_INFO __VA_ARGS__)
#else /*--- #if defined(AVM_WATCHDOG_DEBUG) ---*/
#define DBG(...)
#endif /*--- #else ---*/ /*--- #if defined(AVM_WATCHDOG_DEBUG) ---*/

#include <asm/arch/io.h>
#include <asm/arch/hardware.h>
#include <asm/arch/hw_timer.h>

volatile struct timer *TIMER2 = (struct timer *)io_p2v(DAVINCI_TIMER2_BASE);
static atomic_t wdt_active;

/*-----------------------------------------------------------------------------------------------*\
\*-----------------------------------------------------------------------------------------------*/
void ar7wdt_hw_init(void)
{
	unsigned int reg;

	TIMER2->tcr.Register = 0;
	TIMER2->tgcr.Register = 0;

	TIMER2->tgcr.Register = 8; /* timer mode 64 Bit Watchdog */
	TIMER2->tgcr.Register = 0xb;

	/* assume clock ist  27 MHz */
	TIMER2->prd12 = 27000000 * 40;
	TIMER2->prd34 = 0;

	TIMER2->wdtcr.Register = 0x4000;
	TIMER2->wdtcr.Register =
		0xDA7E4000; /*--- watchdog to activate state ---*/

	atomic_set(&wdt_active, 1);
	return;
}
/**--------------------------------------------------------------------------------**\
\**--------------------------------------------------------------------------------**/
int ar7wdt_hw_is_wdt_running(void)
{
	return atomic_read(&wdt_active);
}
/**--------------------------------------------------------------------------------**\
 * dummy
\**--------------------------------------------------------------------------------**/
void ar7wdt_hw_secure_wdt_disable(void)
{
}
/*-----------------------------------------------------------------------------------------------*\
\*-----------------------------------------------------------------------------------------------*/
void ar7wdt_hw_deinit(void)
{
	TIMER2->prd12 = (unsigned int)-1;
	TIMER2->prd34 = (unsigned int)-1;
	atomic_set(&wdt_active, 0);
	DBG("stopping WDT (wait for long)\n");
	return;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void davinci_watchdog_reset(void)
{
	printk("reboot by watchdog hardware ...\n");
	/* illegal if watchdog enable */
	TIMER2->wdtcr.Register = 0x00004000;
	TIMER2->wdtcr.Register =
		0xA5C64000; /*--- watchdog to pre-activate state ---*/
	TIMER2->wdtcr.Register = 0x00004000;
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void ar7wdt_hw_reboot(void)
{
	panic("[ar7wdt_hw_reboot] soft reboot\n");
}

/*------------------------------------------------------------------------------------------*\
\*------------------------------------------------------------------------------------------*/
void ar7wdt_hw_trigger(void)
{
	if (!ar7wdt_hw_is_wdt_running()) {
		return;
	}
	TIMER2->wdtcr.Register =
		0xA5C64000; /*--- watchdog to pre-activate state ---*/
	TIMER2->wdtcr.Register =
		0xDA7E4000; /*--- watchdog to activate state ---*/
}
EXPORT_SYMBOL(ar7wdt_hw_trigger);
