/*
 * Driver for EIP97 cryptographic accelerator.
 *
 * Copyright (c) 2018 AVM GmbH <fritzbox_info@avm.de>
 *
 * SPDX-License-Identifier: GPL-2.0-only
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This driver is heavily based on the Mediatek driver from Ryder Lee.
 *
 */

#ifndef __EIP97_UTEST_H_
#define __EIP97_UTEST_H_

#include <linux/crypto.h>
#include <crypto/aead.h>

#define MAX_CHUNKS 16

struct template_attr {
	const u8 *data;
	u32 len;
	u32 chunks[MAX_CHUNKS];
	u32 misaligns[MAX_CHUNKS];
	u32 nchunks;
};

enum template_dir { ENCRYPT, DECRYPT };

struct ablkcipher_template {
	struct template_attr key;
	struct template_attr in;
	struct template_attr out;
	struct template_attr iv;
	unsigned int nbytes;
	enum template_dir dir;
};

struct aead_template {
	struct template_attr key;
	struct template_attr in;
	struct template_attr out;
	struct template_attr iv;
	unsigned int textlen;
	unsigned int assoclen;
	unsigned int taglen;
	enum template_dir dir;
};

struct test_template {
	const char *cra_name;
	union {
		struct ablkcipher_template ablkcipher;
		struct aead_template aead;
	} attrs;
};

extern struct test_template tests[];

void eip97_aead_request(struct crypto_aead *tfm,
			struct aead_template *test_data,
			struct aead_alg *alg);

void eip97_test_ahash(struct crypto_alg *alg);
void eip97_test_aead(struct aead_alg *alg);


#endif /* __EIP97_UTEST_H_ */
