#include <linux/kernel.h>
#include <linux/avm_kernel_config.h>
#include <asm/prom.h>

struct _avm_kernel_version_info *avm_kernel_version_info;
struct _avm_kernel_module_memory_config *avm_kernel_module_memory_config;

static inline struct _avm_kernel_config *get_avm_kernel_config_ptr(void)
{
	extern unsigned int __avm_kernel_config_start __attribute__((weak));

	if (IS_ERR(&__avm_kernel_config_start) ||
	    (&__avm_kernel_config_start == NULL))
		return NULL;

	return (struct _avm_kernel_config *)__avm_kernel_config_start;
}

void init_avm_kernel_config(void)
{
	const char *intro = "AVM Kernel Config";
	struct _avm_kernel_config *p;


	p = get_avm_kernel_config_ptr();
	pr_err("[%s] %s (ptr %p)\n", __func__, intro, p);

	if (p == NULL)
		return;

	while (p->tag <= avm_kernel_config_tags_last) {
		if (p->config == NULL)
			return;

		switch (p->tag) {
		case avm_kernel_config_tags_undef:
			pr_err("[%s] %s: undef entry\n", __func__, intro);
			break;
		case avm_kernel_config_tags_module_memory:
			pr_err("[%s] %s: module memory entry\n", __func__,
			       intro);
			avm_kernel_module_memory_config =
				(struct _avm_kernel_module_memory_config
					 *)(p->config);
			break;
		case avm_kernel_config_tags_version_info:
			pr_err("[%s] %s: version info entry\n", __func__,
			       intro);
			avm_kernel_version_info =
				(struct _avm_kernel_version_info *)(p->config);
			break;
		case avm_kernel_config_tags_last:
			return;
		case avm_kernel_config_tags_avmnet:
			pr_err("[%s] %s: unhandled avmnet entry\n", __func__,
			       intro);
			break;
		case avm_kernel_config_tags_hw_config:
			pr_err("[%s] %s: unhandled hw_config entry\n", __func__,
			       intro);
			break;
		case avm_kernel_config_tags_cache_config:
			pr_err("[%s] %s: unhandled cache_config entry\n",
			       __func__, intro);
			break;
		case avm_kernel_config_tags_urlader_env:
			pr_err("[%s] %s: unhandled urlader environment entry\n", __func__,
			       intro);
			break;
		case avm_kernel_config_tags_device_tree_subrev_0:
#ifdef CONFIG_AVM_DT_FROM_CONFIG
			/*--- The Urlader makes sure, that the correct Device Tree is stored at the position of subrevision 0 ---*/
			pr_err("[%s] %s: device-tree found\n", __func__, intro);
			fw_passed_dtb = ((unsigned long)p->config + 0x00UL);
#else
			pr_err("[%s] %s: unhandled device-tree for subrev 0 found\n", __func__, intro);
#endif
			break;
		case avm_kernel_config_tags_device_tree_subrev_1 ... avm_kernel_config_tags_device_tree_subrev_last:
			pr_err("[%s] %s: unhandled device-tree for subrev %d found\n", __func__, intro,
					p->tag - avm_kernel_config_tags_device_tree_subrev_0);
			break;
		}
		p++;
	}

	pr_err("[%s] %s: internal error, should not be reached.\n", __func__, intro);
}

/* vim: set noexpandtab sw=8 ts=8 sts=0: */
